/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * Simple-minded MCF5282 Queued ADC support
 * Non-multiplexed scanning operation.
 */

#include <bsp.h>
#include <epicsStdioRedirect.h>
#include <cantProceed.h>
#include <asynInt32.h>
#include <epicsExport.h>

#define NADC    8

/*
 * Per-driver information
 */
struct dpvt {
    asynInterface                 common;
    asynInterface                 int32;
};

/*
 * asynCommon methods
 */
static asynStatus
connectQADC(void *drvPvt, asynUser *pasynUser)
{
    pasynManager->exceptionConnect(pasynUser);
    return asynSuccess;
}

static asynStatus
disconnectQADC(void *drvPvt, asynUser *pasynUser)
{
    pasynManager->exceptionDisconnect(pasynUser);
    return asynSuccess;
}

static void
reportQADC(void *drvPvt, FILE *fp, int level)
{
    if (level >= 1) {
        fprintf(fp, "QASR0:%2.2X  QASR1:%2.2X\n", MCF5282_QADC_QASR0, MCF5282_QADC_QASR1);
        if (level >= 2) {
            int i;
            fprintf(fp, "QADCMCR:%4.4X  QUACR0:%4.4X  QUACR1:%4.4X  QUACR2:%4.4X\n",
                                MCF5282_QADC_QADCMCR, MCF5282_QADC_QACR0,
                                MCF5282_QADC_QACR1, MCF5282_QADC_QACR2);
            for (i = 0 ; i < NADC ; i++) {
                fprintf(fp, " RJURR[%d]:%3.3X\n", i, MCF5282_QADC_RJURR(i));
            }
            if (level >= 3) {
                for (i = 0 ; i <= NADC ; i++) {
                    fprintf(fp, " CCW[%d]:%3.3X\n", i, MCF5282_QADC_CCW(i));
                }
            }
        }
    }
}

static struct asynCommon commonMethods = { reportQADC,
                                           connectQADC,
                                           disconnectQADC };


/*
 * asynInt32 methods
 */
static asynStatus
getBoundsQADC(void *pvt, asynUser *pasynUser, epicsInt32 *plow, epicsInt32 *phigh)
{
    /*
     * The high value is actually (2^10)-1, but reporting that causes
     * problems with unexpected EOFF values.
     * It's easier to report a symmetrical range here and make sure that
     * EGUF reflects this fact.
     */
    *plow = 0;
    *phigh = 1UL << 10;
    return asynSuccess;
}

static asynStatus
readQADC(void *pvt, asynUser *pasynUser, epicsInt32 *value)
{
    int addr;

   if (pasynManager->getAddr(pasynUser, &addr) != asynSuccess)
       return asynError;
    if ((addr < 0) || (addr >= NADC))
       return asynError;
    *value = MCF5282_QADC_RJURR(addr);
    return asynSuccess;
}

static asynInt32 int32Methods = { NULL,
                                  readQADC,
                                  getBoundsQADC };

/*
 * start up the ADC
 */
static void
initializeQADC()
{
    int i;
    int qpr;
    char *portName = "QADC";
    struct dpvt *dpvt;
    static const unsigned char chanMap[NADC] = { 0, 1, 2, 3, 52, 53, 55, 56 };

    /*
     * Register with ASYN
     */
    dpvt = callocMustSucceed(sizeof(*dpvt), 1, "initializeQADC");
    if (pasynManager->registerPort(portName, ASYN_MULTIDEVICE, 1, 0, 0) != asynSuccess) {
        printf("registerPort failed\n");
        return;
    }
    dpvt->common.interfaceType = asynCommonType;
    dpvt->common.pinterface  = &commonMethods;
    dpvt->common.drvPvt = dpvt;
    if (pasynManager->registerInterface(portName, &dpvt->common) != asynSuccess) {
        printf("registerInterface failed\n");
        return;
    }
    dpvt->int32.interfaceType = asynInt32Type;
    dpvt->int32.pinterface  = &int32Methods;
    dpvt->int32.drvPvt = dpvt;
    if (pasynManager->registerInterface(portName, &dpvt->int32) !=asynSuccess) {
        printf("registerInterface failed\n");
        return;
    }

    /*
     * Enable QADC operation
     */
    MCF5282_QADC_QADCMCR = 0;

    /*
     * Enable all analog inputs
     */
    MCF5282_QADC_DDRQA = 0;
    MCF5282_QADC_DDRQB = 0;

    /*
     * Set channel command words
     */
    for (i = 0 ; i < NADC ; i++)
        MCF5282_QADC_CCW(i) = MCF5282_QADC_CCW_CHAN(chanMap[i]);
    MCF5282_QADC_CCW(i) = MCF5282_QADC_CCW_CHAN(63);

    /*
     * Conversion clock <= 2 MHz
     * Periodic timer (2^17 clocks, about 15.25 Hz), continuous scan
     * Everything from queue 2
     */
    qpr = ((bsp_get_CPU_clock_speed() + 3999999) / (2 * 2000000)) - 1;
    MCF5282_QADC_QACR0 = MCF5282_QADC_QACR0_QPR(qpr);
    MCF5282_QADC_QACR1 = MCF5282_QADC_QACRx_MQ(0);
    MCF5282_QADC_QACR2 = MCF5282_QADC_QACRx_MQ(0x1E) |
                         MCF5282_QADC_QACRx_BQ(0);
}

epicsExportRegistrar(initializeQADC);
